"""This module contains a class for representing the gameboard."""
from random import randint
from collections import namedtuple
import json
import pandas as pd
from .constants import Constants as C
from .gameboard_delegate import GameBoardDelegate
from .combiners import combine_left, combine_right
from .combiners import combine_left_mapped, combine_right_mapped


# a named tuple for the history points of a gameboard
GameBoardSnapshot = namedtuple('GameBoardSnapshot',
                               'board_bits board_json score move')


class GameBoard(object):
    """
    This class models a 2048 game board with a table of tiles holding numbers.

    The board is modeled by a 64 bit integer and all tile, row, column lookups
    are performed using bit-shift operations.
    """

    # the board to put the tiles on in row-major ordering this means the first
    #  index refers to the row and the second index to the column.
    _board = 0

    # the score of the game
    _score = 0

    # the history of metadata throughout the game
    _history = []

    # the delegate to send events to
    _delegate = None

    def __init__(self, delegate=None):
        """
        Create a new gameboard.

        - parameters:
          - delegate: {GameBoardEventHandler} the delegate to pass events to.
        """
        # reset the board (i.e. score and board set to 0)
        self.reset()
        # assign the delegate
        if delegate and not isinstance(delegate, GameBoardDelegate):
            msg = 'delegate must be of type None or GameBoardDelegate'
            raise ValueError(msg)
        self._delegate = delegate

    @property
    def copy(self):
        """Return a copy of this object."""
        copy = GameBoard()
        # copy the board from the other board, use the private setters, no
        # need to run the verification logic as the board should always be
        # in a valid state when copy is called and copy should be fast.
        copy._board = self.board
        # copy the score
        copy._score = self.score
        # setup the move map
        # copy._move_map = self._move_map
        return copy

    def reset(self):
        """Reset the board and score."""
        # reset the score
        self._score = 0
        # create the tile objects by iterating over the array
        self._board = 0
        # restore the history
        self._history = []

    @property
    def history(self):
        """Return the history of this gameboard."""
        return self._history

    @property
    def data_frame(self):
        """
        Return the history as a `pandas` DataFrame.

        - returns: {pandas.DataFrame} the history as a data frame
        """
        columns = ['64-bit board', 'JSON board', 'score', 'move']
        return pd.DataFrame.from_records(self._history, columns=columns)

    @property
    def board(self):
        """Return the board encoded as a 64 bit unsigned int."""
        return self._board

    @property
    def board_json(self):
        """Return the board encoded into json."""
        rows = []
        for column in range(0, C.columns()):
            rows.append(self.get_row_array(column))
        return json.dumps({'board': rows})

    @board.setter
    def board(self, new_board):
        """
        Set the board to a new value.

        - parameters:
            - new_board: {long} the new board to set
        """
        self._board = new_board

    @property
    def board_array(self):
        """Return the board as a 2D array of bytes."""
        board = []
        for row in range(0, C.rows()):
            # add a column list to the array
            board.append([])
            for column in range(0, C.columns()):
                # append the tile in the column to the array
                board[row].append(self.get_tile(row, column))
        return board

    @property
    def score(self):
        """Return the current score of the game."""
        return self._score

    @score.setter
    def score(self, new_score):
        """
        Set the score to a new value.

        - parameters:
            - new_score: {number} the new score for the game, must be positive.
        """
        if new_score < 0:
            raise ValueError('scores cannot be negative')
        self._score = new_score

    def is_equal(self, other):
        """
        Return a boolean determining the equality of this board to the given.

        - parameters:
            - other: {GameBoard} the gameboard to compare against
        - returns: {boolean} true if the boards are equal, false otherwise
        """
        # if the boards are the same, there was no change
        if self.board == other.board and self.score == other.score:
            return False
        return True

    def get_tile(self, row, column):
        """
        Get a tile at a given row column index.

        - parameters:
            - row: {int} the row to get the tile from
            - column: {int} the column to get the tile from
        - return: {int} the tile as a nibble
        """
        # verify the bound of the row
        if row < 0 or row >= C.rows():
            msg = 'row must be in [0, %i)'
            raise ValueError(msg % C.rows())
        # verify the bounds of the column
        if column < 0 or column >= C.columns():
            msg = 'column must be in [0, %i)'
            raise ValueError(msg % C.columns())
        # get the tiles unwrapped index
        index = (row * C.rows()) + column
        return self.get_tile_at(index)

    def get_tile_at(self, index):
        """
        Get the tile at the given index in the bitmap.

        - parameters:
            - index: {number} the index of the tile to get
        - returns: {int} a nibble representing a single tiles exponent value
        """
        # verify the bound of the row
        if index < 0 or index >= C.tiles():
            msg = 'index must be in [0, %s)'
            raise ValueError(msg % C.tiles())
        # get the offset based on index
        row_offset = (C.rows() * index)
        offset = C.board_size() - C.tile_size() - row_offset
        # shift the board and mask it to get the value
        return (self._board >> offset) & C.tile_value_mask()

    def place_tile(self, row, column, value):
        """
        Place a tile at the given row column index with the given value.

        - parameters:
            - row: {int} the index of the row to place at
            - column: {int} the index of the column to place at
            - value: {int} the value to place in [0, 15]
        """
        # verify the bounds of the row
        if row < 0 or row >= C.rows():
            msg = 'row must be in [0, %i)'
            raise ValueError(msg % C.columns())
        # verify the bounds of the column
        if column < 0 or column >= C.columns():
            msg = 'column must be in [0, %i)'
            raise ValueError(msg % C.columns())
        # get the index
        index = (row * C.rows()) + column
        # place the tile
        self.place_tile_at(index, value)

    def place_tile_at(self, index, value):
        """
        Place a new tile at the unwrapped index.

        - parameters:
            - index: {int} the index to place the value at
            - value: {long} the value to place in the tile
        """
        # verify the bound of the index
        if index < 0 or index >= C.tiles():
            msg = 'tile must be in [0, %i)'
            raise ValueError(msg % C.tiles())
        if value < 0 or value > 15:
            raise ValueError('value must be in [0, 15]')
        # calculate the offset
        offset = C.rows() * index
        # shift the zero to delete the old value
        self._board = self._board & ~(C.tile_mask() >> offset)
        # zero out garbage and shift value for placement, then add the new
        # value to the board
        mask = C.board_size() - C.tile_size() - offset
        masked_value = ((value & C.tile_value_mask()) << mask)
        self._board = self._board + masked_value

    def place_random_tile(self):
        """Randomly place a new tile on the board."""
        # make sure there is a place for the tile
        if len(self.active_tiles) is C.tiles():
            return
        # get a list of the inactive tiles
        inactive_tiles = self.inactive_tiles
        value = 1  # 2 ^ 1 = 2
        probability = randint(1, 1.0 / C.chance_of_four())
        if probability is 1:
            value = 2  # 2 ^ 2 = 4
        # get a random index in the list
        index = randint(0, len(inactive_tiles) - 1)
        # place the tile on the board
        self.place_tile_at(inactive_tiles[index], value)
        # pass the event to the handler
        if self._delegate is not None:
            tile_index = inactive_tiles[index]
            row = tile_index / C.rows()
            column = tile_index % C.rows()
            # self._delegate.didPlaceNewTile(row, column, value)

    def place_row(self, index, value):
        """
        Place a row onto the board.

        - parameters:
            - index: {int} the index of the row to replace
            - value: {short} the value of the row to put into the board
        """
        # cross out the row to make room for the new one
        self._board = self._board & ~(C.row_mask() >> int(C.row_size() * index))
        # cast the row and mask any sign extension, shift the row, then add the row to the board
        self._board = self._board + (((value) & C.short_mask()) << int((C.rows() - 1 - index) * C.row_size()))

    def get_row(self, index):
        """
        Return the encoded row value for a given index.

        - parameters:
            - index: {int} the index of the row to select [0, 3]
        - returns: {short} the row as 16 bits such that each 4 represents a
                           tile
        """
        shifted_board = self.board >> int((C.rows() - 1 - index) * C.row_size())
        return shifted_board & C.short_mask()

    def place_row_array(self, index, row_array):
        """
        Place a row array on the board.

        - parameters:
            - index: {int} the index of the row to place
            - row_array: {byte[]} the list of tiles
        """
        for column in range(0, C.columns()):
            self.place_tile(index, column, row_array[column])

    def get_row_array(self, row):
        """
        Get the tiles in a row as an array of bytes.

        - parameters:
            - row: {int} the index of the row to get a tile array of
        - returns: {byte[]} an array of tiles
        """
        row_tiles = []
        for column in range(0, C.columns()):
            row_tiles.append(self.get_tile(row, column))
        return row_tiles

    def place_column(self, index, value):
        """
        Place a column on the board.

        - parameters:
            - index: {int} the index to place the column at
            - value: {short} the value of the column to place
        """
        # shift the mask to the given column, flip the bits, then & with the board to zero the
        # column.
        self._board = self._board & ~(C.column_mask() >> int(C.tile_size() * index))
        # first get the pieces of the board shifted into column positions, then shift the column
        # itself into place and add it to the board.
        self._board = self._board + (
            (
                # get the pieces of the board shifted into column positions as a long
                ( ((value >> C.tile_shift()[0]) & C.tile_value_mask()) << C.column_shift()[0] ) +
                ( ((value >> C.tile_shift()[1]) & C.tile_value_mask()) << C.column_shift()[1] ) +
                ( ((value >> C.tile_shift()[2]) & C.tile_value_mask()) << C.column_shift()[2] ) +
                (   value & C.tile_value_mask()                     )
                #then shift the column
            ) << (C.rows() - 1 - index) * C.tile_size()
        )

    def get_column(self, index):
        """
        Get the encoded column value for a given column.

        - parameters:
            - index: {int} the index of the column to select [0, 3]
        - returns: {short} the column as 16 bits such that each 4 represents a
                           tile
        """
        return ((self.get_tile(0, index) << C.tile_shift()[0]) +
                (self.get_tile(1, index) << C.tile_shift()[1]) +
                (self.get_tile(2, index) << C.tile_shift()[2]) +
                (self.get_tile(3, index)))

    def place_column_array(self, index, column_array):
        """
        Place a column array on the board.

        - parameters:
            - index: {int} the index of the column to place
            - column_array: {byte[]} the list of tiles
        """
        for row in range(0, C.rows()):
            self.place_tile(row, index, column_array[row])

    def get_column_array(self, column):
        """
        Get the tiles in a column as an array of bytes.

        - parameters:
            - column: {int} the index of the row to get a tile array of
        - returns: {byte[]} an array of tiles
        """
        column_tiles = []
        # iterate over the rows
        for row in range(0, C.rows()):
            column_tiles.append(self.get_tile(row, column))
        return column_tiles

    @property
    def inactive_tiles(self):
        """Return a list containing pointers to the inactive tiles."""
        inactive_tiles = []

        for index in range(0, C.tiles()):
            if self.get_tile_at(index) is 0:
                inactive_tiles.append(index)

        return inactive_tiles

    @property
    def active_tiles(self):
        """Return a list containing pointers to the active tiles."""
        active_tiles = []

        for index in range(0, C.tiles()):
            if self.get_tile_at(index) is not 0:
                active_tiles.append(index)

        return active_tiles

    # MARK: Movement

    def _move(self, length, combine, get, place):
        """
        Perform a move function dynamically on a row / col array.

        - parameters:
            - length: {int} the number of row / col to perform a move action on
            - combine: {function} a combine function that returns a new board
                                  and score as a tuple
            - get: {function} the function that will return a row / col
            - place: {function} the function that will place a new row / col
        """
        for index in range(0, length):
            combination = combine(get(index))
            place(index, combination[0])
            self._score = self._score + combination[1]

    def _move_up(self):
        """
        Move the pieces up on the game board and increment the score.

        - precondition: assume the move is legal
        """
        self._move(C.rows(), combine_left,
                   self.get_column_array, self.place_column_array)

    def _move_down(self):
        """
        Move the pieces down on the game board and increment the score.

        - precondition: assume the move is legal
        """
        self._move(C.rows(), combine_right,
                   self.get_column_array, self.place_column_array)

    def _move_left(self):
        """
        Move the pieces left on the game board and increment the score.

        - precondition: assume the move is legal
        """
        self._move(C.columns(), combine_left,
                   self.get_row_array, self.place_row_array)

    def _move_right(self):
        """
        Move the pieces right on the game board and increment the score.

        - precondition: assume the move is legal
        """
        self._move(C.columns(), combine_right,
                   self.get_row_array, self.place_row_array)

    # MARK: Can move

    @property
    def can_move_up(self):
        """Return a boolean determining whether a move up is possible."""
        copy = self.copy
        copy._move_up()
        return self.is_equal(copy)

    @property
    def can_move_down(self):
        """Return a boolean determining whether a move down is possible."""
        copy = self.copy
        copy._move_down()
        return self.is_equal(copy)

    @property
    def can_move_left(self):
        """Return a boolean determining whether a move left is possible."""
        copy = self.copy
        copy._move_left()
        return self.is_equal(copy)

    @property
    def can_move_right(self):
        """Return a boolean determining whether a move right is possible."""
        copy = self.copy
        copy._move_right()
        return self.is_equal(copy)

    @property
    def can_move(self):
        """Return a boolean determining if there are any moves on the board."""
        return self.can_move_up or self.can_move_down or self.can_move_left or self.can_move_right

    # Guarded Movement

    def _try_move(self, can_move, move, did_move=None):
        if can_move:
            # get the matrix of moves
            # int[][] moveMatrix = LikeValueIndexMapper.getUpMovementMatrixFor(self)
            # store the old score (for the handler)
            old_score = self.score
            # make the move
            move()
            # take a snapshot
            snapshot = GameBoardSnapshot(board_bits=self._board,
                                         board_json=self.board_json,
                                         score=self._score,
                                         move=str(move.__name__))
            self._history.append(snapshot)
            # if there is a handler, pass the event
            if did_move:
                did_score = self.score > old_score
                # did_move(did_score, moveMatrix)
            return True
        return False

    def move_up(self):
        """
        Try to move the pieces up on the game board.

        If the  move isn't possible then return false and leave the game in
        its current state.
        - returns: {boolean} true if the move occured, false otherwise
        """
        return self._try_move(self.can_move_up, self._move_up)

    def move_down(self):
        """
        Try to move the pieces down on the game board.

        If the  move isn't possible then return false and leave the game in
        its current state.
        - returns: {boolean} true if the move occured, false otherwise
        """
        return self._try_move(self.can_move_down, self._move_down)

    def move_left(self):
        """
        Try to move the pieces left on the game board.

        If the  move isn't possible then return false and leave the game in
        its current state.
        - returns: {boolean} true if the move occured, false otherwise
        """
        return self._try_move(self.can_move_left, self._move_left)

    def move_right(self):
        """
        Try to move the pieces right on the game board.

        If the  move isn't possible then return false and leave the game in
        its current state.
        - returns: {boolean} true if the move occured, false otherwise
        """
        return self._try_move(self.can_move_right, self._move_right)

    # MARK: Description

    @property
    def description(self):
        """Return a human readable version of this gameboard."""
        description = ""
        # iterate over all the tiles
        for i in range(0, C.tiles()):
            # get the tile at this index
            value = self.get_tile_at(i)
            if value is 0:
                tile = '|{0: >6} |'.format('')
            else:
                tile = '|{0: >6} |'.format(2**value)
            # add the tile to the string
            description += tile
            # if the index is about to be a multiple of four, add a new line
            if (i + 1) % 4 is 0:
                description += "\n"
        return description
